@@ -40,17 +40,53 @@ from PyQt5.QtCore import (
     QSettings)
 from PyQt5.QtCore import pyqtSignal as Signal, pyqtSlot as Slot
 
-#from ..gui.utils import message_warning, message_information, \
-#                        message_critical as message_error, \
-#                        OSX_NSURL_toLocalFile
-#from ..help.manager import get_dist_meta, trim, parse_meta
+from urllib.request import urlopen
 
 from orangecanvas.gui.utils import message_warning, message_information, \
                         message_critical as message_error
 from orangecanvas.help.manager import get_dist_meta, trim, parse_meta
 
+from orangecanvas.resources import package_dirname
 
 PYPI_API_JSON = "https://pypi.org/pypi/{name}/json"
+
+#####################################################
+# MAKE INTERNAL LIBRARIES DYNAMICAL
+
+# read add-on list
+
+INTERNAL_LIBRARIES = [a.strip() for a in open(os.path.join(package_dirname("oasys.application"), "data", "INTERNAL_LIBRARIES.txt"), "rt")]
+INTERNAL_LIBRARIES = [a for a in INTERNAL_LIBRARIES if a]
+OFFICIAL_ADDONS = [a.strip() for a in open(os.path.join(package_dirname("oasys.application"), "data", "OFFICIAL_ADDONS.txt"), "rt")]
+OFFICIAL_ADDONS = [a for a in OFFICIAL_ADDONS if a]
+
+# query PyPI
+
+internal_libraries_list = []
+is_auto_update = True
+
+try:
+    for package in INTERNAL_LIBRARIES:
+        r = urlopen(PYPI_API_JSON.format(name=package)).read().decode("utf-8")
+        p = json.loads(r)
+        p["releases"] = p["releases"][p["info"]["version"]] # load only the last version
+
+        internal_libraries_list.append(p)
+except:
+    is_auto_update = False
+
+official_addons_list = []
+
+try:
+    for package in OFFICIAL_ADDONS:
+        r = urlopen(PYPI_API_JSON.format(name=package)).read().decode("utf-8")
+        p = json.loads(r)
+        p["releases"] = p["releases"][p["info"]["version"]] # load only the last version
+
+        official_addons_list.append(p)
+except:
+    is_auto_update = False
+
 OFFICIAL_ADDON_LIST = "https://raw.githubusercontent.com/oasys-kit/oasys-addons/master/list"
 OFFICIAL_ADDON_LIST_ALTERNATIVE = "https://rawcdn.githack.com/oasys-kit/oasys-addons/91dbd16c78f2ce42f4abe65e72c17abe064e0520/list"
 
@@ -455,7 +491,7 @@ class AddonManagerDialog(QDialog):
         addmore.clicked.connect(self.__run_add_package_dialog)
 
         buttons.accepted.connect(self.__accepted)
-        buttons.rejected.connect(self.reject)
+        buttons.rejected.connect(self.__rejected)
 
         empty = QWidget()
         empty.setFixedHeight(1)
@@ -481,6 +517,11 @@ class AddonManagerDialog(QDialog):
         if not self._f_pypi_addons.done():
             self.__progressDialog()
 
+        self.__is_app_to_be_closed = False
+
+    def set_is_app_to_be_closed(self, is_app_to_be_closed=True):
+        self.__is_app_to_be_closed = is_app_to_be_closed
+
     def __run_add_package_dialog(self):
         dlg = QDialog(self, windowTitle="Add add-on by name")
         dlg.setAttribute(Qt.WA_DeleteOnClose)
@@ -664,7 +705,6 @@ class AddonManagerDialog(QDialog):
 
     ADDON_EXTENSIONS = ('.zip', '.whl', '.tar.gz')
 
-
     def dragEnterEvent(self, event):
         urls = event.mimeData().urls()
         if any((OSX_NSURL_toLocalFile(url) or url.toLocalFile())
@@ -690,6 +730,13 @@ class AddonManagerDialog(QDialog):
         self._set_packages(future)
         self.addonwidget.set_install_projects(names)
 
+    def __rejected(self):
+        self.reject()
+        if self.__is_app_to_be_closed:
+            message = "Click Ok to restart OASYS for changes to take effect."
+            message_information(message, parent=self)
+            sys.exit(0)
+
     def __accepted(self):
         steps = self.addonwidget.item_state()
 
@@ -731,14 +778,43 @@ class AddonManagerDialog(QDialog):
         self.accept()
         sys.exit(0)
 
+
+import platform
+
+def list_available_internal_libraries():
+    if is_auto_update:
+
+        system, node, release, version, machine, processor = platform.uname()
+
+        packages = []
+        for library in internal_libraries_list:
+            try:
+                info = library["info"]
+
+                if not ("Debian 3" in version and info["name"]=="PyQt5"):
+                    packages.append(
+                    Installable(info["name"], info["version"],
+                                info["summary"], info["description"],
+                                info["package_url"],
+                                info["package_url"])
+                )
+            except (TypeError, KeyError):
+                continue  # skip invalid packages
+
+        return packages
+
 def list_available_versions():
     """
     List add-ons available.
     """
-    try:
-        addons = requests.get(OFFICIAL_ADDON_LIST).json()
-    except:
-        addons = requests.get(OFFICIAL_ADDON_LIST_ALTERNATIVE).json()
+
+    if is_auto_update:
+        addons = official_addons_list
+    else:
+        try:
+            addons = requests.get(OFFICIAL_ADDON_LIST).json()
+        except:
+            addons = requests.get(OFFICIAL_ADDON_LIST_ALTERNATIVE).json()
 
     # query pypi.org for installed add-ons that are not in our list
     installed = list_installed_addons()
@@ -823,9 +899,7 @@ def installable_from_json_response(meta):
 def list_installed_addons():
     from oasys.canvas.conf import ADDONS_ENTRY
     workingset = pkg_resources.WorkingSet(sys.path)
-    return [ep.dist for ep in
-            workingset.iter_entry_points(ADDONS_ENTRY)]
-
+    return [ep.dist for ep in workingset.iter_entry_points(ADDONS_ENTRY)]
 
 def unique(iterable):
     seen = set()
@@ -896,18 +970,26 @@ def installable_items(pypipackages, installed=[]):
         items.append(item)
     return items
 
+def update_internal_libraries(internal_libraries_to_update=None):
+    try:
+        pip = PipInstaller()
+
+        if not internal_libraries_to_update is None:
+            to_update_list = [item.installable for item in internal_libraries_to_update]
+        else:
+            to_update_list = list_available_internal_libraries()
+
+        for installable in to_update_list:
+            pip.upgrade(installable)
+            print("Updated: " + installable.name)
+    except Exception as ex:
+        print("Internal libraries update failed")
 
 Install, Upgrade, Uninstall = 1, 2, 3
 
+from oasys.util.external_command import CommandFailed, run_command
 
-class CommandFailed(Exception):
-    def __init__(self, cmd, retcode, output):
-        if not isinstance(cmd, str):
-            cmd = " ".join(map(shlex.quote, cmd))
-        self.cmd = cmd
-        self.retcode = retcode
-        self.output = output
-
+IS_WINDOW = sys.platform == "win32"
 
 class Installer(QObject):
     installStatusChanged = Signal(str)
@@ -915,12 +997,13 @@ class Installer(QObject):
     finished = Signal()
     error = Signal(str, object, int, list)
 
-    def __init__(self, parent=None, steps=[]):
+    def __init__(self, parent=None, steps=[], use_conda = False):
         QObject.__init__(self, parent)
         self.__interupt = False
         self.__queue = deque(steps)
         self.pip = PipInstaller()
-        self.conda = CondaInstaller()
+        if use_conda: self.conda = CondaInstaller()
+        else: self.conda = None
 
     def start(self):
         QTimer.singleShot(0, self._next)
@@ -939,15 +1022,14 @@ class Installer(QObject):
             if command == Install:
                 self.setStatusMessage(
                     "Installing {}".format(cleanup(pkg.installable.name)))
-                if self.conda:
-                    self.conda.install(pkg.installable, raise_on_fail=False)
-                self.pip.install(pkg.installable)
+                if self.conda: self.conda.install(pkg.installable, raise_on_fail=False)
+                self.pip.install(pkg.installable, deep=not IS_WINDOW)
             elif command == Upgrade:
                 self.setStatusMessage(
                     "Upgrading {}".format(cleanup(pkg.installable.name)))
                 if self.conda:
                     self.conda.upgrade(pkg.installable, raise_on_fail=False)
-                self.pip.upgrade(pkg.installable)
+                self.pip.upgrade(pkg.installable, deep=not IS_WINDOW)
             elif command == Uninstall:
                 self.setStatusMessage(
                     "Uninstalling {}".format(cleanup(pkg.local.project_name)))
@@ -977,8 +1059,10 @@ class PipInstaller:
         arguments = QSettings().value('add-ons/pip-install-arguments', '', type=str)
         self.arguments = shlex.split(arguments)
 
-    def install(self, pkg):
-        cmd = ["python", "-m", "pip", "install"]
+    def install(self, pkg, deep=True):
+        if deep: cmd = ["python", "-m", "pip", "install"]
+        else:    cmd = ["python", "-m", "pip", "install", "--user"]
+
         cmd.extend(self.arguments)
         if pkg.package_url.startswith("http://") or pkg.package_url.startswith("https://"):
             cmd.append(pkg.name)
@@ -988,14 +1072,16 @@ class PipInstaller:
 
         run_command(cmd)
 
-    def upgrade(self, package):
+    def upgrade(self, package, deep=True):
         # This is done in two steps to avoid upgrading
         # all of the dependencies - faster
-        self.upgrade_no_deps(package)
+        self.upgrade_no_deps(package, deep)
+
         self.install(package)
 
-    def upgrade_no_deps(self, package):
-        cmd = ["python", "-m", "pip", "install", "--upgrade", "--no-cache-dir"]
+    def upgrade_no_deps(self, package, deep=True):
+        if deep: cmd = ["python", "-m", "pip", "install", "--upgrade", "--no-cache-dir"]
+        else:    cmd = ["python", "-m", "pip", "install", "--upgrade", "--user", "--no-cache-dir"]
         cmd.extend(self.arguments)
         cmd.append(package.name)
 
@@ -1008,8 +1094,8 @@ class PipInstaller:
 
 class CondaInstaller:
     def __init__(self):
-        enabled = QSettings().value('add-ons/allow-conda',
-                                    True, type=bool)
+        enabled = QSettings().value('add-ons/allow-conda', True, type=bool)
+
         if enabled:
             self.conda = self._find_conda()
         else:
@@ -1057,82 +1143,6 @@ class CondaInstaller:
         return bool(self.conda)
 
 
-def run_command(command, raise_on_fail=True):
-    """Run command in a subprocess.
-
-    Return `process` return code and output once it completes.
-    """
-    log.info("Running %s", " ".join(command))
-
-    if command[0] == "python":
-        process = python_process(command[1:])
-    else:
-        process = create_process(command)
-
-    output = []
-    while process.poll() is None:
-        try:
-            line = process.stdout.readline()
-        except IOError as ex:
-            if ex.errno != errno.EINTR:
-                raise
-        else:
-            output.append(line)
-            print(line, end="")
-    # Read remaining output if any
-    line = process.stdout.read()
-    if line:
-        output.append(line)
-        print(line, end="")
-
-    if process.returncode != 0:
-        log.info("Command %s failed with %s",
-                 " ".join(command), process.returncode)
-        log.debug("Output:\n%s", "\n".join(output))
-        if raise_on_fail:
-            raise CommandFailed(command, process.returncode, output)
-
-    return process.returncode, output
-
-
-def python_process(args, script_name=None, **kwargs):
-    """
-    Run a `sys.executable` in a subprocess with `args`.
-    """
-    executable = sys.executable
-    if os.name == "nt" and os.path.basename(executable) == "pythonw.exe":
-        # Don't run the script with a 'gui' (detached) process.
-        dirname = os.path.dirname(executable)
-        executable = os.path.join(dirname, "python.exe")
-
-    if script_name is not None:
-        script = script_name
-    else:
-        script = executable
-
-    return create_process(
-        [script] + args,
-        executable=executable
-    )
-
-
-def create_process(cmd, executable=None, **kwargs):
-    if hasattr(subprocess, "STARTUPINFO"):
-        # do not open a new console window for command on windows
-        startupinfo = subprocess.STARTUPINFO()
-        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
-        kwargs["startupinfo"] = startupinfo
-
-    return subprocess.Popen(
-        cmd,
-        executable=executable,
-        cwd=None,
-        stderr=subprocess.STDOUT,
-        stdout=subprocess.PIPE,
-        bufsize=-1,
-        universal_newlines=True,
-        **kwargs
-    )
 
 from PyQt5.QtCore import QPointF, QUrl
 
